from abc import ABC

from super_scad_thread.lead_thread.ThreadLeadCreator import ThreadLeadCreator


class InternalThreadLeadCreator(ThreadLeadCreator, ABC):
    """
    Creates a lead thread on a 2D internal thread profile.
    """
    # Nothing to implement.
    pass

# ----------------------------------------------------------------------------------------------------------------------
